@@ -175,6 +175,7 @@ struct supplicant_result {
 struct supplicant_task {
 	int ifindex;
 	char *ifname;
+	gboolean mac80211;
 	struct connman_device *device;
 	struct connman_network *network;
 	struct connman_network *pending_network;
@@ -533,7 +534,6 @@ static int remove_interface(struct supplicant_task *task)
 	return -EINPROGRESS;
 }
 
-#if 0
 static int set_ap_scan(struct supplicant_task *task)
 {
 	DBusMessage *message, *reply;
@@ -572,7 +572,6 @@ static int set_ap_scan(struct supplicant_task *task)
 
 	return 0;
 }
-#endif
 
 static int add_network(struct supplicant_task *task)
 {
@@ -715,18 +714,15 @@ static int select_network(struct supplicant_task *task)
 	return 0;
 }
 
-static int enable_network(struct supplicant_task *task)
+static int disconnect_network(struct supplicant_task *task)
 {
 	DBusMessage *message, *reply;
 	DBusError error;
 
 	DBG("task %p", task);
 
-	if (task->netpath == NULL)
-		return -EINVAL;
-
-	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->netpath,
-					SUPPLICANT_INTF ".Network", "enable");
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
+				SUPPLICANT_INTF ".Interface", "disconnect");
 	if (message == NULL)
 		return -ENOMEM;
 
@@ -741,45 +737,7 @@ static int enable_network(struct supplicant_task *task)
 			connman_error("%s", error.message);
 			dbus_error_free(&error);
 		} else
-			connman_error("Failed to enable network");
-		dbus_message_unref(message);
-		return -EIO;
-	}
-
-	dbus_message_unref(message);
-
-	dbus_message_unref(reply);
-
-	return 0;
-}
-
-static int disable_network(struct supplicant_task *task)
-{
-	DBusMessage *message, *reply;
-	DBusError error;
-
-	DBG("task %p", task);
-
-	if (task->netpath == NULL)
-		return -EINVAL;
-
-	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->netpath,
-					SUPPLICANT_INTF ".Network", "disable");
-	if (message == NULL)
-		return -ENOMEM;
-
-	dbus_message_set_auto_start(message, FALSE);
-
-	dbus_error_init(&error);
-
-	reply = dbus_connection_send_with_reply_and_block(connection,
-							message, -1, &error);
-	if (reply == NULL) {
-		if (dbus_error_is_set(&error) == TRUE) {
-			connman_error("%s", error.message);
-			dbus_error_free(&error);
-		} else
-			connman_error("Failed to disable network");
+			connman_error("Failed to disconnect network");
 		dbus_message_unref(message);
 		return -EIO;
 	}
@@ -841,9 +799,12 @@ static int set_network(struct supplicant_task *task,
 						DBUS_TYPE_STRING, &passphrase);
 	} else if (g_ascii_strcasecmp(security, "wep") == 0) {
 		const char *key_mgmt = "NONE";
-		const char *auth_alg = "OPEN SHARED";
+		const char *auth_alg = "OPEN";
 		const char *key_index = "0";
 
+		if (task->mac80211 == TRUE)
+			auth_alg = "OPEN SHARED";
+
 		connman_dbus_dict_append_variant(&dict, "auth_alg",
 						DBUS_TYPE_STRING, &auth_alg);
 
@@ -1199,7 +1160,6 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	unsigned short channel, frequency;
 	const char *mode, *security;
 	char *group = NULL;
-	unsigned int ssid_len;
 
 	DBG("task %p", task);
 
@@ -1337,10 +1297,8 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	if (result.name != NULL && result.name[0] != '\0')
 		connman_network_set_name(network, result.name);
 
-	if (connman_network_get_blob(network, "WiFi.SSID", &ssid_len) == NULL) {
-		connman_network_set_blob(network, "WiFi.SSID",
-					 result.ssid, result.ssid_len);
-	}
+	connman_network_set_blob(network, "WiFi.SSID",
+						result.ssid, result.ssid_len);
 
 	connman_network_set_string(network, "WiFi.Mode", mode);
 
@@ -1565,20 +1523,42 @@ static int task_connect(struct supplicant_task *task)
 	if (g_str_equal(security, "none") == FALSE && passphrase == NULL)
 		return -EINVAL;
 
-	add_network(task);
+	remove_network(task);
 
-	select_network(task);
-	disable_network(task);
+	set_ap_scan(task);
+
+	add_network(task);
 
 	set_network(task, ssid, ssid_len, address, security, passphrase);
 
-	err = enable_network(task);
+	err = select_network(task);
 	if (err < 0)
 		return err;
 
 	return -EINPROGRESS;
 }
 
+static void scanning(struct supplicant_task *task, DBusMessage *msg)
+{
+	DBusError error;
+	dbus_bool_t scanning;
+
+	dbus_error_init(&error);
+
+	if (dbus_message_get_args(msg, &error, DBUS_TYPE_BOOLEAN, &scanning,
+						DBUS_TYPE_INVALID) == FALSE) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Wrong arguments for scanning");
+		return;
+	}
+
+	connman_info("%s scanning %s", task->ifname,
+				scanning == TRUE ? "started" : "finished");
+}
+
 static void state_change(struct supplicant_task *task, DBusMessage *msg)
 {
 	DBusError error;
@@ -1602,6 +1582,8 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 
 	DBG("state %s ==> %s", oldstate, newstate);
 
+	connman_info("%s %s", task->ifname, newstate);
+
 	state = string2state(newstate);
 	if (state == WPA_INVALID)
 		return;
@@ -1619,7 +1601,7 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 	switch (task->state) {
 	case WPA_COMPLETED:
 		if (get_bssid(task->device, bssid, &bssid_len) == 0)
-	                connman_network_set_address(task->network,
+			connman_network_set_address(task->network,
 							bssid, bssid_len);
 
 		/* carrier on */
@@ -1627,7 +1609,7 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 		break;
 
 	case WPA_DISCONNECTED:
-		disable_network(task);
+		remove_network(task);
 
 		/* carrier off */
 		connman_network_set_connected(task->network, FALSE);
@@ -1640,7 +1622,8 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 				task->network = task->pending_network;
 				task->pending_network = NULL;
 				task_connect(task);
-			}
+			} else
+				task->network = NULL;
 		}
 		break;
 
@@ -1659,7 +1642,8 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 				task->network = task->pending_network;
 				task->pending_network = NULL;
 				task_connect(task);
-			}
+			} else
+				task->network = NULL;
 		}
 		break;
 
@@ -1695,6 +1679,8 @@ static DBusHandlerResult supplicant_filter(DBusConnection *conn,
 
 	if (g_str_equal(member, "ScanResultsAvailable") == TRUE)
 		scan_results_available(task);
+	else if (g_str_equal(member, "Scanning") == TRUE)
+		scanning(task, msg);
 	else if (g_str_equal(member, "StateChange") == TRUE)
 		state_change(task, msg);
 
@@ -1720,6 +1706,10 @@ int supplicant_start(struct connman_device *device)
 		goto failed;
 	}
 
+	task->mac80211 = connman_inet_is_mac80211(task->ifindex);
+	if (task->mac80211 == FALSE)
+		connman_warn("Enabling quirks for unsupported driver");
+
 	task->range = g_try_malloc0(sizeof(struct iw_range));
 	if (task->range == NULL) {
 		err = -ENOMEM;
@@ -1778,10 +1768,10 @@ int supplicant_stop(struct connman_device *device)
 	if (task->scanning == TRUE)
 		connman_device_set_scanning(task->device, FALSE);
 
-	disable_network(task);
-
 	remove_network(task);
 
+	disconnect_network(task);
+
 	return remove_interface(task);
 }
 
@@ -1864,10 +1854,10 @@ int supplicant_disconnect(struct connman_network *network)
 	if (task->disconnecting == TRUE)
 		return -EALREADY;
 
-	disable_network(task);
-
 	remove_network(task);
 
+	disconnect_network(task);
+
 	task->disconnecting = TRUE;
 
 	return 0;
@@ -1986,10 +1976,7 @@ int supplicant_register(struct supplicant_driver *driver)
 
 	driver_list = g_slist_append(driver_list, driver);
 
-	if (g_dbus_check_service(connection, SUPPLICANT_NAME) == TRUE)
-		supplicant_probe(connection, NULL);
-	else
-		supplicant_activate(connection);
+	supplicant_activate(connection);
 
 	return 0;
 }
